# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

import numpy as np

from itertools import izip
from itertools import tee

from threedigrid.orm.models import Model
from threedigrid.orm.fields import ArrayField
from threedigrid.orm.fields import PointArrayField
from threedigrid.orm.fields import PolygonArrayField
from threedigrid.orm.utils import transform_xys
from threedigrid.admin.utils import get_smallest_uint_dtype
from threedigrid.admin.nodes import exporters
from threedigrid.admin.nodes import subsets


def pairwise(iterable):
    # from https://docs.python.org/2/library/
    # itertools.html#recipes
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return izip(a, b)


NODE_SUBSETS = {
    'node_type__eq': subsets.NODE_TYPE__EQ_SUBSETS,
    'node_type__in': subsets.NODE_TYPE__IN_SUBSETS,
}


class Nodes(Model):
    content_pk = ArrayField()
    seq_id = ArrayField()
    coordinates = PointArrayField()
    cell_coords = PolygonArrayField()
    zoom_category = ArrayField()
    node_type = ArrayField()
    SUBSETS = NODE_SUBSETS

    @property
    def connectionnodes(self):
        return self._filter_as(ConnectionNodes, content_pk__ne=0)

    @property
    def manholes(self):
        return self._filter_as(Manholes, content_pk__ne=0)

    @property
    def added_calculationnodes(self):
        return self._filter_as(AddedCalculationNodes, content_pk=0)

    def __init__(self, *args, **kwargs):

        super(Nodes, self).__init__(*args, **kwargs)

        self._exporters = [
            exporters.NodesOgrExporter(self),
        ]

    @property
    def locations_2d(self):
        data = self.subset('2D_open_water').data
        # x0 = 0, y0 = 0
        # Translate
        # data['coordinates'][0] += x0
        # data['coordinates'][1] += y0

        # Return [[node_id, coordinate_x + x0, coordinate_y + y0]]
        return zip(
            data['id'].tolist(),
            data['coordinates'][0].tolist(),
            data['coordinates'][1].tolist())


class AddedCalculationNodes(Nodes):
    pass


class ConnectionNodes(Nodes):
    initial_waterlevel = ArrayField()
    storage_area = ArrayField()


class Manholes(ConnectionNodes):
    bottom_level = ArrayField()
    drain_level = ArrayField()
    display_name = ArrayField()
    surface_level = ArrayField()
    calculation_type = ArrayField()
    shape = ArrayField()
    width = ArrayField()
    manhole_indicator = ArrayField()


class Cells(Nodes):

    z_coordinate = ArrayField()

    def __init__(self, *args, **kwargs):

        super(Cells, self).__init__(*args, **kwargs)

        self._exporters = [
            exporters.CellsOgrExporter(self),
        ]

    @property
    def bounds(self):
        """
        :return: coordinates of the cell bounds (counter clockwise):
                 minx, miny, maxx, miny, maxx, maxy, minx, maxy
        """
        minx, miny, maxx, maxy = self.cell_coords
        return np.vstack(
            (minx, miny,
             maxx, miny,
             maxx, maxy,
             minx, maxy))

    def get_id_from_xy(self, x, y, xy_epsg_code=None, subset_name=None):
        """
        :param x: the x coordinate in xy_epsg_code
        :param y: the y coordinate in xy_epsg_code
        :param subset_name: filter on a subset of cells

        :return: numpy array with cell id's for x, y
        """

        if xy_epsg_code and xy_epsg_code != self.epsg_code:
            xy = transform_xys(
                np.array([x]), np.array([y]),
                xy_epsg_code, self.epsg_code).flatten()
        else:
            xy = [x, y]

        inst = self
        if subset_name:
            inst = self.subset(subset_name)
        id = inst.filter(cell_coords__contains_point=xy).id
        if id.size == 1:
            return int(id[0])
        return None

    def __repr__(self):
        return "<orm cells instance of {}>".format(self.model_name)

    def __contenttype__(self):
        return 'cells'


class Grid(Model):
    """nodm, nodn and nodk have the same size as nodes and cells
    which is why this model lives in the nodes/model module. In fact
    they are attributes of the cell coordinates"""

    nodm = ArrayField()  #
    nodn = ArrayField()
    nodk = ArrayField()

    def __init__(self, *args, **kwargs):

        super(Grid, self).__init__(*args, **kwargs)
        self.n2dtot = kwargs['n2dtot']
        self.dx = kwargs['dx']

    def get_pixel_map(self, dem_pixelsize, dem_shape):
        """
        get the node grid to pixel map

        :param dem_pixelsize: pixelsize of the geo tiff
        dem_shape: shape of the numpy representation of the geo tiff
        """

        # Convert nod_grid to smallest uint type possible
        dtype = get_smallest_uint_dtype(maxval=self.n2dtot)
        grid_arr = np.zeros(dem_shape, dtype=dtype)

        # applies for for 2D nodes only
        _k = self.nodk[0:self.n2dtot + 1] - 1
        _m = self.nodm[0:self.n2dtot + 1] - 1
        _n = self.nodn[0:self.n2dtot + 1] - 1

        # the size in pixels of each grid cell
        _size = self.dx[_k] / dem_pixelsize
        # corresponding node index
        _ind = np.arange(0, self.n2dtot + 1, dtype='int')

        n_slice_start = np.array(_n * _size, dtype='int')
        n_slice_end = np.array((_n * _size) + _size, dtype='int')
        n_slice = zip(n_slice_start, n_slice_end)
        m_slice_start = np.array(_m * _size, dtype='int')
        m_slice_end = np.array((_m * _size) + _size, dtype='int')
        m_slice = zip(m_slice_start, m_slice_end)

        for ns, ms, idx in zip(n_slice, m_slice, _ind):
            axis_x = slice(*ns)
            axis_y = slice(*ms)
            grid_arr[axis_x, axis_y] = idx

        # flip upside-down to match geotiff
        return grid_arr[::-1, ::]



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

import numpy as np
import logging
from itertools import izip
from itertools import tee
from itertools import chain

from abc import ABCMeta
from collections import OrderedDict

from threedigrid.orm.base.exceptions import OperationNotSupportedError
from threedigrid.orm.base.fields import ArrayField
from threedigrid.orm.base.fields import IndexArrayField
from threedigrid.orm.base.filters import get_filter
from threedigrid.orm.base.filters import SliceFilter

logger = logging.getLogger(__name__)


def pairwise(iterable):
    # from https://docs.python.org/2/library/
    # itertools.html#recipes
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return izip(a, b)


class Model:
    __metaclass__ = ABCMeta
    id = IndexArrayField()

    _field_names = []

    _exporters = []

    _filter_map = None

    _datasource = None

    def __init__(self, datasource=None, slice_filters=[],
                 epsg_code=None, only_fields=[], reproject_to_epsg=None,
                 has_1d=None, **kwargs):
        """
        Initialize a Model with a datasource, filters
        and a epsg_code.

        The datasource is a wrapped h5py Group from a H5py file or
        an instance that implements the same interface.
        """
        self._datasource = datasource
        self.starts_at = 0  # 0 or 1 based array
        self.slice_filters = slice_filters
        self.only_fields = only_fields
        self.reproject_to_epsg = reproject_to_epsg

        if not epsg_code:
            epsg_code = self._datasource.getattr('epsg_code')
        self.epsg_code = epsg_code

        # Cache the field names
        self._field_names = [
            x for x in dir(self.__class__)
            if isinstance(getattr(self.__class__, x), ArrayField)]
        self.has_1d = has_1d
        self.model_name = '-'.join(
            (self._datasource.getattr('model_name'),
             self._datasource.getattr('model_slug'),
             str(self._datasource.getattr('revision_nr')),
             self._datasource.getattr('revision_hash')))

    def get_field(self, field_name):
        """
        Returns: the ArrayField with field_name on this instance.
        """
        return super(Model, self).__getattribute__(field_name)

    def get_field_value(self, field_name):
        """
        Returns: the value for the ArrayField with name field_name
        on the instance.

        With this in place we can later override the ArrayField
        get_value method for field specific conversion, if needed.
        """
        return self.get_field(field_name).get_value(
            self._datasource, field_name)

    def __getattribute__(self, attr_name):
        """
        Override the __getattribute__ methode to return
        the computed values of ArrayField's instead of
        the ArrayField
        instance.
        """

        # Note: don't use getattr(self, XX) in this function,
        # it will lead to a max recursion error
        attr = super(Model, self).__getattribute__(attr_name)
        # Override getting ArrayField attributes
        if isinstance(attr, ArrayField):

            # Return meta directly, we don't want it to
            # to be filtered
            if attr_name == 'meta':
                return attr.get_value(self._datasource, attr_name)

            # Use the get function to retrieve the computed/filtered
            # value for the ArrayField with name: 'name'
            get_func = super(Model, self).__getattribute__('to_dict')
            return get_func()[attr_name]

        # Default behaviour, return the attribute from superclass
        return attr

    @property
    def fields(self):
        """
        Returns: a list of ArrayFields names (excluding 'meta')
        """
        return [x for x in self._field_names if x != 'meta']

    def __init_class(self, klass, slice_filters, only_fields=[],
                     reproject_to_epsg=None):
        """
        Returns: a new instance of 'klass' with new filters and
                 same datasource.
        """
        return klass(
            datasource=self._datasource,
            slice_filters=slice_filters,
            only_fields=only_fields,
            epsg_code=self.epsg_code,
            reproject_to_epsg=reproject_to_epsg,
            has_1d=self.has_1d)

    def __get_filters(self, **kwargs):
        """
        Parse kwargs for filters.

        For example:
            content_type__in=[1,2,3]

            splitted_key[0] = "content_type"
            splitted_key[1] = "in"
            -> Try to find filter with "in" as key

        """
        new_slice_filters = list(self.slice_filters)  # make copy

        for key, value in kwargs.iteritems():
            splitted_key = key.split('__')
            if splitted_key[0] not in self.fields:
                raise ValueError(
                    "Field '{}' unknown. Choices are {}".format(
                        splitted_key[0],
                        self.fields)
                )

            new_slice_filters.append(
                get_filter(
                    splitted_key,
                    self.get_field(splitted_key[0]),
                    value,
                    filter_map=self._filter_map)
            )

        return new_slice_filters

    def _filter_as(self, klass, **kwargs):
        """
        Same as self.filter, but now return "klass" instance instead of
        self.__class__ instance
        """
        slice_filters = self.__get_filters(**kwargs)
        return self.__init_class(klass, slice_filters)

    def filter(self, **kwargs):
        """
        Django style filtering, allows multiple filter or chaining filters
        for example:

            line_instance.filter(content_type='v2_weir', content_pk=1)
            line_instance.filter(content_type='v2_weir').filter(content_pk=1)

            Both produce the same result object.


        Returns: a new instance of the object with the extra filters added.
        """
        slice_filters = self.__get_filters(**kwargs)
        return self.__init_class(self.__class__, slice_filters)

    def slice(self, s, override_filter_error=False):
        """
        slice by name or slice. See instance.predefined_slices for an overview
        of predefined slices

        :param s: either the name of the slice (see instance.predefined_slices
             for an overview or an slice instance, e.g. slice(<x>, <y>)
        :return: nodes instance with containing the sliced dataset.Example
            usage::
            usage:
            open_water_2d = nodes.slice('2D_open_water')
            open_water_2d.data

                OrderedDict([('content_pk', array([1, 2, 3, ...

        """
        if self.slice_filters and not override_filter_error:
            raise OperationNotSupportedError(
                'You cannot use slices on a filtered dataset')

        slice_filter = s

        if not isinstance(s, slice):
            raise TypeError(
                'Type %s not supported, must be a slice' % type(s,)
            )

        return self.__init_class(
            self.__class__, [SliceFilter(slice_filter)] + self.slice_filters)

    @property
    def has_groundwater(self):
        req = self.subset('2D_GROUNDWATER').data
        return len(req['id']) > 0

    @property
    def known_subset(self):
        if not hasattr(self, 'SUBSETS'):
            return "has no subsets defined"
        return list(
            chain(*[v.keys() for v in self.SUBSETS.itervalues()])
        )

    def subset(self, name):
        """
        get a subset of the data. See instance.subsets for an overview
        of predefined subsets

        :param name: name of the subset (see instance.predefined_slices
             for an overview or an slice instance, e.g. slice(<x>, <y>)
        :return: nodes instance with containing the sliced dataset.Example
            usage::
            open_water_2d = nodes.slice('2D_open_water')
            open_water_2d.data

                OrderedDict([('content_pk', array([1, 2, 3, ...

        """
        if not hasattr(self, 'SUBSETS'):
            raise TypeError("SUBSETS not defined for this type of model")

        if isinstance(name, basestring):
            field_filter = [key for key in self.SUBSETS if
                            name.upper() in self.SUBSETS[key]]

            if not field_filter:
                msg = 'Invalid subset name.'
                msg += ' Valid options are: %s ' % (
                    [x for y in self.SUBSETS.values() for x in y], )
                logger.exception(msg)
                raise KeyError(msg)

            if len(field_filter) > 1:
                raise KeyError(
                    'Multiple options found for subset: %s ' % name.upper())

            field_filter = field_filter[0]
        else:
            raise TypeError('Type %s not supported' % type(name,))

        return self.filter(
            **{field_filter: self.SUBSETS[field_filter][name.upper()]})

    def only(self, *args):
        """
        Return only the fields requested

        Usage: line_instance.only('content_pk', 'line')
        """
        if not args:
            raise Exception("Please provide at least one field name")

        for x in args:
            if x not in self.fields:
                raise Exception("Unknown field name: {0}".format(x))

        return self.__init_class(
            self.__class__, self.slice_filters,
            only_fields=self.only_fields + list(args))

    def __do_filter(self):
        """
        Performs: the filters defined in self._filters

        Returns: a dictionairy with filtered values
        """

        # No cached data
        # filter all the data using the defined filters.

        selection = OrderedDict()

        for n in self.fields:
            selection[n] = self.get_field_value(n)

        # Apply all filters sequential on the selection dict
        # Every filter can, when matches are found, remove items
        # from the values in the dict.
        #
        # For example filtering on content_pk=2 works like:
        #      1. Get the boolean mask:
        #              mask = selection['content_pk'][
        #                  selection['content_pk'] ==  2]
        #      2. Apply it to all values in selection.
        #              selection[x] = selection[x][mask]
        #              (for x in selection.keys())
        for filter_instance in self.slice_filters:
            filter_instance.filter_dict(selection, self)

        # Reproject any coordinates if a reproject_to_epsg is set and
        # there are coordinatefields in the selection
        if self.reproject_to_epsg and self._includes_coords(selection):
            selection = self.__do_reproject(
                    selection, self.reproject_to_epsg)

        # Prune all unwanted fields
        if self.only_fields:
            for key in [x for x in selection.keys()
                        if x not in self.only_fields]:
                selection.pop(key)
        if not self.slice_filters:
            selection = self._get_values(selection)
        return selection

    def _get_values(self, selection):
        """
        when no filters are specified, we are still operating on the hdf5
        groups so get the values explicitly

        :param selection: ordered dict of selected fields
            with group datasets
        :return: new ordered dict with values from group datasets
        """

        _tmp = OrderedDict()
        for k, v in selection.iteritems():
            try:
                _tmp[k] = v[:]
            except TypeError:
                pass
        return _tmp

    def to_dict(self):
        """
        Returns: the filtered values as dictionairy
        """
        # Note: the __do_filter functions returns
        # a dictionairy by default.
        return self.__do_filter()

    def to_list(self):
        """
        :return: list of dicts with key's and values
        """

        # Filter results and transform the result to
        # and np.ndarray
        selection = self.__do_filter()
        if len(selection.values()) > 1:
            array = np.array(selection.values())
        else:
            array = selection.values()[0]

        def optional_zip(array_to_zip):
            """
            Optional zip the results of
            array_to_zip if there are more than one
            dimensions.
            """
            # Transform data to list
            array_as_list = array_to_zip.tolist()
            if len(array_to_zip.shape) > 1:
                # Zip the list if there are more than
                # one dimensions.
                array_as_list = zip(*array_as_list)
            return array_as_list

        # Zip the lists of array_lists (does same as a transpose)
        data = zip(*[optional_zip(x) for x in array])

        # Create a list of dictionairies of the data by
        # zipping the selection.keys() (= field_names) for all items
        # in data
        return [dict(zip(selection.keys(), x)) for x in data]

    def to_array(self):
        """
        Returns: the filtered values as a numpy array
        """
        selection = self.__do_filter()
        if len(selection.values()) > 1:
            return np.array(selection.values())
        return selection.values()[0]

    @property
    def data(self):
        """
        Returns: the (filtered) data a dictionairy
        """
        return self.to_dict()

    def __repr__(self):
        """human readable representation of the instance"""
        return "<orm {} instance of {}>".format(
            self.__contenttype__(), self.model_name
        )

    def __contenttype__(self):
        """conent type, e.g. lines, nodes, cells, ..."""
        return self._datasource.group_name



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

import logging
from threedigrid.orm.base.datasource import DataSource


logger = logging.getLogger(__name__)


class H5pyGroup(DataSource):
    """
    Datasource wrapper for h5py groups,
    adding meta data to all groups.
    """
    _group_name = None
    _h5py_file = None

    def __init__(self, h5py_file, group_name, meta=None, required=False):
        if group_name not in h5py_file.keys() and not required:
            logger.info(
                '[*] {0} not found in file {1}, not required...'.format(
                    group_name, h5py_file)
            )
            return

        self.group_name = group_name

        self._h5py_file = h5py_file
        try:
            self._source = h5py_file.require_group(group_name)
        except TypeError:
            self._source = h5py_file[group_name]

        self.meta = dict(
            [(y, [x.value]) for y, x in self._h5py_file.get('meta').items()])
        self.meta['trash'] = [1]

    def set(self, name, values):
        if name in self._source:
            # Overwrite
            self._source[name][:] = values
        else:
            # Create
            self._source.create_dataset(name, data=values)

    def getattr(self, name):
        return self._h5py_file.attrs[name]


    def keys(self):
        return self._source.keys()



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function


class ArrayField:
    """
    Generic field that can be used to describe values
    to be retrieved from a Datasource.
    """
    @staticmethod
    def get_value(datasource, name):
        """
        Returns: the data from the datasource
                 or None if 'name' is not in the datasource

        Optional transforms can be done here.
        """
        if name in datasource.keys():
            return datasource[name]

        return None


class IndexArrayField(ArrayField):
    def __init__(self, to=None):
        self.to = to



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import unittest
import tempfile
import shutil

import ogr

from threedigrid.admin.gridadmin import GridH5Admin

from threedigrid.admin.lines.exporters import LinesOgrExporter


test_file_dir = os.path.join(
    os.getcwd(), "tests/test_files")

# the testfile is a copy of the v2_bergermeer gridadmin file
grid_admin_h5_file = os.path.join(test_file_dir, "gridadmin.h5")


class ExporterTestShp(unittest.TestCase):

    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)
        d = tempfile.mkdtemp()
        self.f_shp = os.path.join(d, "exporter_test_lines.shp")

    def tearDown(self):
        shutil.rmtree(os.path.dirname(self.f_shp))

    def test_export_by_extension(self):
        line_2d_open_water_wgs84 = self.parser.lines.subset(
            '2D_OPEN_WATER').reproject_to('4326')
        exporter = LinesOgrExporter(line_2d_open_water_wgs84)
        exporter.save(self.f_shp, line_2d_open_water_wgs84.data, '4326')
        self.assertTrue(os.path.exists(self.f_shp))
        s = ogr.Open(self.f_shp)
        layer = s.GetLayer()
        self.assertEqual(
            layer.GetFeatureCount(), line_2d_open_water_wgs84.id.size)


class ExporterTestGpkg(unittest.TestCase):

    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)
        d = tempfile.mkdtemp()
        self.f_gpkg = os.path.join(d, "exporter_test_lines.gpkg")

    def tearDown(self):
        shutil.rmtree(os.path.dirname(self.f_gpkg))

    def test_export_by_extension(self):
        line_2d_open_water_wgs84 = self.parser.lines.subset(
            '2D_OPEN_WATER').reproject_to('4326')
        exporter = LinesOgrExporter(line_2d_open_water_wgs84)
        exporter.save(self.f_gpkg, line_2d_open_water_wgs84.data, '4326')
        self.assertTrue(os.path.exists(self.f_gpkg))
        s = ogr.Open(self.f_gpkg)
        layer = s.GetLayer()
        self.assertEqual(
            layer.GetFeatureCount(), line_2d_open_water_wgs84.id.size)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

import numpy as np

from threedigrid.admin.lines import exporters
from threedigrid.orm.models import Model
from threedigrid.orm.fields import (
    ArrayField, LineArrayField, MultiLineArrayField)
from threedigrid.orm.base.fields import IndexArrayField
from threedigrid.admin.lines import subsets


LINE_SUBSETS = {
    'kcu__in': subsets.KCU__IN_SUBSETS
}


class Lines(Model):
    kcu = ArrayField()
    lik = ArrayField()
    line = IndexArrayField(to='Nodes')
    content_pk = ArrayField()
    content_type = ArrayField()
    zoom_category = ArrayField()
    line_coords = LineArrayField()
    line_geometries = MultiLineArrayField()
    SUBSETS = LINE_SUBSETS

    def __init__(self, *args,  **kwargs):
        super(Lines, self).__init__(*args, **kwargs)

        self._exporters = [
            exporters.LinesOgrExporter(self),
        ]

    @property
    def v2_channels(self):
        return self.filter(content_type='v2_channel')

    @property
    def pipes(self):
        return self._filter_as(Pipes, content_type='v2_pipe')

    @property
    def weirs(self):
        return self._filter_as(Weirs, content_type='v2_weir')

    @property
    def orifices(self):
        return self._filter_as(Orifices, content_type='v2_orifice')

    @property
    def channels(self):
        return self._filter_as(Channels, content_type='v2_channel')

    @property
    def culverts(self):
        return self._filter_as(Culverts, content_type='v2_culvert')

    @property
    def line_nodes(self):
        return np.dstack(self.line)[0]


class Pipes(Lines):
    display_name = ArrayField()
    invert_level_start_point = ArrayField()
    invert_level_end_point = ArrayField()
    friction_type = ArrayField()
    friction_value = ArrayField()
    cross_section_height = ArrayField()
    cross_section_width = ArrayField()
    cross_section_shape = ArrayField()
    sewerage_type = ArrayField()
    calculation_type = ArrayField()
    connection_node_start_pk = ArrayField()
    connection_node_end_pk = ArrayField()


class Channels(Lines):
    code = ArrayField()
    calculation_type = ArrayField()
    dist_calc_points = ArrayField()
    connection_node_start_pk = ArrayField()
    connection_node_end_pk = ArrayField()


class Weirs(Lines):
    code = ArrayField()
    display_name = ArrayField()
    discharge_coefficient_negative = ArrayField()
    discharge_coefficient_positive = ArrayField()
    sewerage = ArrayField()
    friction_type = ArrayField()
    friction_value = ArrayField()
    crest_type = ArrayField()
    crest_level = ArrayField()
    connection_node_start_pk = ArrayField()
    connection_node_end_pk = ArrayField()

    @property
    def line_coord_angles(self):
        # Filter raw values with content_type and content_pk
        # to always get the raw (metric) values
        # and not the reprojected ones to calculate
        # the angles with.
        content_type = self._datasource['content_type'].value
        line_content_pk = self._datasource['content_pk'].value
        pk_mask = np.isin(line_content_pk, self.content_pk)
        raw_values = self._datasource['line_coords'].value[
            :, (content_type == 'v2_weir') & pk_mask]

        return self.get_field('line_coords').get_angles_in_degrees(
            raw_values)


class Culverts(Lines):
    display_name = ArrayField()
    code = ArrayField()
    calculation_type = ArrayField()
    dist_calc_points = ArrayField()
    cross_section_height = ArrayField()
    cross_section_width = ArrayField()
    cross_section_shape = ArrayField()
    friction_type = ArrayField()
    friction_value = ArrayField()
    discharge_coefficient_negative = ArrayField()
    discharge_coefficient_positive = ArrayField()
    invert_level_start_point = ArrayField()
    invert_level_end_point = ArrayField()
    connection_node_start_pk = ArrayField()
    connection_node_end_pk = ArrayField()


class Orifices(Lines):
    display_name = ArrayField()
    sewerage = ArrayField()
    max_capacity = ArrayField()
    friction_type = ArrayField()
    friction_value = ArrayField()
    crest_type = ArrayField()
    crest_level = ArrayField()
    discharge_coefficient_negative = ArrayField()
    discharge_coefficient_positive = ArrayField()
    connection_node_start_pk = ArrayField()
    connection_node_end_pk = ArrayField()



import unittest

import numpy as np

from threedigrid.orm.base.filters import BaseFilter
from threedigrid.orm.base.filters import BaseCompareFilter
from threedigrid.orm.base.filters import EqualsFilter
from threedigrid.orm.base.filters import get_filter
from threedigrid.orm.base.filters import InFilter
from threedigrid.orm.filters import PointFilter
from threedigrid.orm.filters import FILTER_MAP

from threedigrid.orm.base.fields import ArrayField
from threedigrid.orm.fields import PolygonArrayField


class BasicFilterTests(unittest.TestCase):

    def setUp(self):
        self.field = ArrayField()

    def test_base_filter(self):
        b = BaseFilter()
        with self.assertRaises(NotImplementedError):
            b.filter('foo')

    def test_repr(self):
        f = BaseCompareFilter('in', self.field, [1, 2, 3])
        self.assertTrue(repr(f))

    def test_get_filter_eq(self):
        """Test that we get a default EqualsFilter without keys."""
        f = get_filter(['content_type'], self.field, 3)
        self.assertIsInstance(f, EqualsFilter)

        # also works when 'eq' is explicitly given
        f = get_filter(['content_type', 'eq'], self.field, 3)
        self.assertIsInstance(f, EqualsFilter)

    def test_eq_filter(self):
        f = EqualsFilter('content_type', self.field, np.array([1, 2]))
        filtered = f.filter({'content_type': np.array([1, 2])})
        self.assertTrue(filtered.all())

        filtered = f.filter({'content_type': np.array([1, 3])})
        self.assertFalse(filtered.all())

    def test_get_filter_in(self):
        f = get_filter(['content_pk', 'in'], self.field, [2])
        self.assertIsInstance(f, InFilter)

    def test_get_filter_error(self):
        """get_filter fails with malformed splitted keys."""
        with self.assertRaises(ValueError):
            get_filter(['content_pk', 'in', 'foo'], self.field, [2])

        with self.assertRaises(ValueError):
            get_filter([], self.field, [2])

    def test_get_nonexisting_filter(self):
        with self.assertRaises(ValueError):
            get_filter(['content_type', 'not_a_filter'], self.field, 3)


class PointFilterTests(unittest.TestCase):

    def setUp(self):
        self.field = PolygonArrayField()

    def test_get_point_filter(self):
        f = get_filter(
            ['cell_coords', 'contains_point'],
            self.field, [1, 5], filter_map=FILTER_MAP
        )
        self.assertIsInstance(f, PointFilter)

    def test_point_filter(self):
        f = PointFilter('coord_cells', self.field, np.array([1, 5]))
        filtered = f.filter(
            {'coord_cells': np.array([[0, 2], [0, 2], [6, 6], [10, 11]])}
        )
        self.assertTrue(filtered[0])
        self.assertFalse(filtered[1])



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

import logging
import os
from threedigrid.admin.h5py_datasource import H5pyGroup
from threedigrid.admin.idmapper import IdMapper
from threedigrid.admin import constants
from threedigrid.admin.gridadmin import GridH5Admin

from threedigrid.admin.breaches.prepare import PrepareBreaches
from threedigrid.admin.lines.prepare import (
    PrepareLines, PrepareChannels, PreparePipes, PrepareWeirs,
    PrepareOrifices, PrepareCulverts)

from threedigrid.admin.levees.prepare import PrepareLevees
from threedigrid.admin.nodes.prepare import (
    PrepareNodes, PrepareManholes, PrepareConnectionNodes)
from threedigrid.admin.pumps.prepare import PreparePumps
from threedigrid.admin.levees.models import Levees

logger = logging.getLogger(__name__)


def is_prepared(h5py_file, group_name, attr_name):
    if group_name not in h5py_file:
        return False
    return h5py_file[group_name].attrs.get(attr_name, 0) == 1


def skip_prepare(h5py_file, group_name, attr_name, overwrite):
    if overwrite:
        return False
    return is_prepared(h5py_file, group_name, attr_name)


def prepare_lines_onedee(
        h5py_file, threedi_datasource, klass, attr_name, overwrite=False):


    if skip_prepare(h5py_file, 'lines', attr_name, overwrite):
        return

    klass.prepare_datasource(h5py_file, threedi_datasource)

    h5py_file['lines'].attrs[attr_name] = 1


class GridAdminH5Prepare(object):
    @staticmethod
    def prepare(h5py_file, threedi_datasource, extra_attrs=None):
        # Step 1: Set optional extra attrs like epsg code,
        # model_name, revision_number, revision_hash
        # and possible other meta data..
        if extra_attrs:
            for key, value in extra_attrs.iteritems():
                h5py_file.attrs[key] = value

        # Step 3: prepare the ID mapper
        IdMapper.prepare_mapper(
            h5py_file, threedi_datasource)

        # Step 4: Prepare nodes/lines
        GridAdminH5Prepare.prepare_nodes(h5py_file, threedi_datasource)
        GridAdminH5Prepare.prepare_lines(h5py_file, threedi_datasource)

        # Step 5: Prepare onedee lines/nodes/pumps
        GridAdminH5Prepare.prepare_onedee_lines(h5py_file, threedi_datasource)
        GridAdminH5Prepare.prepare_onedee_nodes(h5py_file, threedi_datasource)
        GridAdminH5Prepare.prepare_pumps(h5py_file, threedi_datasource)

        # Step 6: prepare levees
        GridAdminH5Prepare.prepare_levees(h5py_file, threedi_datasource)

        # Step 7: prepare breaches
        GridAdminH5Prepare.prepare_breaches(h5py_file, threedi_datasource)

    @staticmethod
    def prepare_levees(h5py_file, threedi_datasource, overwrite=False):
        if skip_prepare(h5py_file, 'levees', 'prepared', overwrite):
            return

        PrepareLevees.prepare_datasource(h5py_file, threedi_datasource)

        h5py_file['levees'].attrs['prepared'] = 1

    @staticmethod
    def prepare_onedee_nodes(h5py_file, threedi_datasource, overwrite=False):
        has_1d = h5py_file.attrs.get('has_1d', 0) == 1

        if not has_1d:
            return

        if not is_prepared(h5py_file, 'nodes', 'prepared'):
            # Line data is missing, prepare line data first.
            GridAdminH5Prepare.prepare_nodes(h5py_file, threedi_datasource)

        if skip_prepare(h5py_file, 'nodes', 'connectionnodes_prepared',
                        overwrite):
            return

        PrepareConnectionNodes.prepare_datasource(
                h5py_file, threedi_datasource)

        h5py_file['nodes'].attrs['connectionnodes_prepared'] = 1

        if skip_prepare(h5py_file, 'nodes', 'manholes_prepared', overwrite):
            return

        PrepareManholes.prepare_datasource(h5py_file, threedi_datasource)

        h5py_file['nodes'].attrs['manholes_prepared'] = 1

    @staticmethod
    def prepare_nodes(h5py_file, threedi_datasource, overwrite=False):
        if skip_prepare(h5py_file, 'nodes', 'prepared', overwrite):
            return

        has_1d = h5py_file.attrs.get('has_1d', 0) == 1

        if has_1d:
            mapping1d = H5pyGroup(h5py_file, 'mapping1d')
        else:
            mapping1d = None

        id_mapper = IdMapper(H5pyGroup(h5py_file, 'mappings')['id_map'])

        # Prepare nodes
        node_group = H5pyGroup(h5py_file, 'nodes')
        PrepareNodes.prepare_datasource(
            node_group, mapping1d, id_mapper, has_1d=has_1d)

        h5py_file['nodes'].attrs['prepared'] = 1

    @staticmethod
    def prepare_lines(h5py_file, threedi_datasource, overwrite=False):
        # Prepare lines
        if skip_prepare(h5py_file, 'lines', 'lines_prepared', overwrite):
            return

        has_1d = h5py_file.attrs.get('has_1d', 0) == 1

        line_group = H5pyGroup(h5py_file, 'lines')
        node_group = H5pyGroup(h5py_file, 'nodes')

        id_mapper = IdMapper(H5pyGroup(h5py_file, 'mappings')['id_map'])

        PrepareLines.prepare_datasource(
            line_group,
            id_mapper, threedi_datasource,
            [node_group['x_coordinate'], node_group['y_coordinate']],
            has_1d=has_1d
        )

        h5py_file['lines'].attrs['lines_prepared'] = 1

    @staticmethod
    def prepare_breaches(h5py_file, threedi_datasource, overwrite=True):
        if skip_prepare(h5py_file, 'breaches', 'prepared', overwrite):
            return

        if not is_prepared(h5py_file, 'lines', 'line_prepared'):
            # Line data is missing, prepare line data first.
            GridAdminH5Prepare.prepare_lines(h5py_file, threedi_datasource)

        if not is_prepared(h5py_file, 'levees', 'prepared'):
            # Levee data is missing, prepare levee data first.
            GridAdminH5Prepare.prepare_levees(h5py_file, threedi_datasource)

        has_breaches = h5py_file.attrs.get('has_breaches', 0) == 1
        levees = None
        if 'levees' in h5py_file:
            levees = Levees(H5pyGroup(h5py_file, 'levees'))

        id_mapper = IdMapper(H5pyGroup(h5py_file, 'mappings')['id_map'])
        line_group = H5pyGroup(h5py_file, 'lines')

        # Prepare breaches
        if has_breaches:
            PrepareBreaches.prepare_datasource(
                H5pyGroup(h5py_file, 'breaches', required=True),
                line_group['kcu'], id_mapper, levees,
                line_group['line_coords']
            )

            h5py_file['breaches'].attrs['prepared'] = 1

    @staticmethod
    def prepare_onedee_lines(h5py_file, threedi_datasource, overwrite=False):
        has_1d = h5py_file.attrs.get('has_1d', 0) == 1

        if not has_1d:
            return

        if not is_prepared(h5py_file, 'lines', 'line_prepared'):
            # Line data is missing, prepare line data first.
            GridAdminH5Prepare.prepare_lines(h5py_file, threedi_datasource)

        # Load data from threedi_datasource into the h5py_file on
        # the lines group.
        prepare_lines_onedee(
            h5py_file, threedi_datasource, PrepareChannels,
            'channels_prepared', overwrite)

        prepare_lines_onedee(
            h5py_file, threedi_datasource, PreparePipes,
            'pipes_prepared', overwrite)

        prepare_lines_onedee(
            h5py_file, threedi_datasource, PrepareWeirs,
            'weirs_prepared', overwrite)

        prepare_lines_onedee(
            h5py_file, threedi_datasource, PrepareOrifices,
            'orifices_prepared', overwrite)

        prepare_lines_onedee(
            h5py_file, threedi_datasource, PrepareCulverts,
            'culverts_prepared', overwrite)

    @staticmethod
    def prepare_pumps(h5py_file, threedi_datasource, overwrite=False):
        has_1d = h5py_file.attrs.get('has_1d', 0) == 1
        has_pumpstations = h5py_file.attrs.get('has_pumpstations', 0) == 1

        if not has_1d or not has_pumpstations:
            return

        if skip_prepare(h5py_file, 'pumps', 'prepared', overwrite):
            return

        # if not h5py_file.attrs.get('has_pumpstations'):
        #    logger.info('[*] Datasource does not have pumps, skipping...')
        #    return

        if not is_prepared(h5py_file, 'nodes', 'prepared'):
            # Node data is missing, prepare node data first.
            GridAdminH5Prepare.prepare_nodes(h5py_file, threedi_datasource)

        PreparePumps.prepare_datasource(
            h5py_file, threedi_datasource, H5pyGroup(h5py_file, 'pumps'))

        h5py_file['pumps'].attrs['prepared'] = 1


class GridAdminH5Export(object):
    """
    Exporter that saves shapefiles to the model's /preprocessed dir
    """

    def __init__(self, gridadmin_file, export_method='to_shape'):
        """

        :param gridadmin_file: hdf5 file (full path)
        """
        self.ga = GridH5Admin(gridadmin_file)
        self._dest = os.path.split(gridadmin_file)[0]
        self._export_method = export_method

    def export_all(self):
        """convenience function to run all exports"""
        self.export_nodes()
        self.export_lines()
        self.export_grid()
        self.export_levees()

    def export_nodes(self):
        """
        writes shapefile of all 1D nodes
        """
        if not self.ga.has_1d:
            logger.info(
                "[*] Model {} does not have 1D, "
                "skipping export nodes...".format(self.ga.model_name)
            )
            return
        dest = os.path.join(self._dest, constants.NODES_SHP)
        getattr(self.ga.nodes.subset(
            constants.SUBSET_1D_ALL), self._export_method)(dest)

    def export_lines(self):
        """
        writes shapefile of all 1D lines
        """
        if not self.ga.has_1d:
            logger.info(
                "[*] Model {} does not have 1D, "
                "skipping export lines...".format(self.ga.model_name)
            )
            return
        dest = os.path.join(self._dest, constants.LINES_SHP)
        getattr(self.ga.lines.subset(
            constants.SUBSET_1D_ALL), self._export_method)(dest)

    def export_levees(self):
        """
        writes shapefile of levees
        """
        if not self.ga.has_levees:
            logger.info(
                "[*] Model {} does not have levees, "
                "skipping export levees...".format(self.ga.model_name)
            )
            return
        dest = os.path.join(self._dest, constants.LEVEES_SHP)
        getattr(self.ga.levees, self._export_method)(dest)

    def export_grid(self):
        """
        writes shapefile of all 2D groundwater and open water
        cells (if present)
        """
        self._export_2d_grid()
        self._export_groundwater_grid()

    def _export_groundwater_grid(self):
        if not self.ga.has_groundwater:
            logger.info(
                "[*] Model {} does not have groundwater, "
                "skipping export groundwater cells...".format(self.ga.model_name)
            )
            return
        dest_gw = os.path.join(self._dest, constants.GROUNDWATER_SHP)
        getattr(self.ga.cells.subset(
            constants.SUBSET_2D_GROUNDWATER), self._export_method)(dest_gw)

    def _export_2d_grid(self):
        if not self.ga.has_2d:
            logger.info(
                "[*] Model {} does not have 2D, "
                "skipping export 2d cells...".format(
                    self.ga.model_name)
            )
            return

        dest_ow = os.path.join(self._dest, constants.OPEN_WATER_SHP)

        getattr(self.ga.cells.subset(
            constants.SUBSET_2D_OPEN_WATER), self._export_method)(dest_ow)



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import unittest
import json

from threedigrid.admin.gridadmin import GridH5Admin
from threedigrid.admin.lines.serializers import ChannelsGeoJsonSerializer
from threedigrid.admin.lines.serializers import WeirsGeoJsonSerializer

test_file_dir = os.path.join(
    os.getcwd(), "tests/test_files")

# the testfile is a copy of the v2_bergermeer gridadmin file
grid_admin_h5_file = os.path.join(test_file_dir, "gridadmin.h5")


class GridAdminSerializerTest(unittest.TestCase):

    def setUp(self):
        self.parser = GridH5Admin(grid_admin_h5_file)

    def test_channel_serializer(self):
        channels_wgs84 = self.parser.lines.channels.subset(
            '1D_ALL').reproject_to('4326')
        channels_wgs84_geojson = ChannelsGeoJsonSerializer(channels_wgs84).data
        j = json.loads(channels_wgs84_geojson)
        feat = j['features']
        self.assertEqual(len(feat), channels_wgs84.id.size)
        self.assertEqual(feat[0]['properties']['object_type'], 'v2_channel')

    def test_weir_serializer(self):
        weirs = self.parser.lines.weirs.subset('1D_ALL').reproject_to('4326')
        wjson = WeirsGeoJsonSerializer(weirs).data
        j = json.loads(wjson)
        feat = j['features']
        self.assertEqual(len(feat), weirs.id.size)
        self.assertEqual(feat[0]['properties']['object_type'], 'v2_weir')



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

import os
import logging
from collections import OrderedDict

from osgeo import ogr

from threedigrid.admin.utils import reshape_flat_array
from threedigrid.admin.utils import get_spatial_reference
from threedigrid.admin.utils import KCUDescriptor
from threedigrid.admin.constants import GEO_PACKAGE_DRIVER_NAME
from threedigrid.admin.constants import OGR_FIELD_TYPE_MAP
from threedigrid.admin.constants import SHP_DRIVER_NAME
from threedigrid.orm.base.exporters import BaseOgrExporter

logger = logging.getLogger(__name__)



class BreachesOgrExporter(BaseOgrExporter):
    def __init__(self, breaches):
        self._breaches = breaches
        self.supported_drivers = {
            SHP_DRIVER_NAME,
            GEO_PACKAGE_DRIVER_NAME,
        }

    def save(self, file_name, breach_data, target_epsg_code, **kwargs):
        assert self.driver is not None
        selection = breach_data
        kcu_dict = KCUDescriptor()
        geomtype = 0
        sr = get_spatial_reference(target_epsg_code)
        self.del_datasource(file_name)
        data_source = self.driver.CreateDataSource(file_name)
        layer = data_source.CreateLayer(
            str(os.path.basename(file_name)),
            sr,
            geomtype
        )
        fields = OrderedDict([
            ('link_id', 'int'),
            ('kcu', 'int'),
            ('kcu_descr', 'str'),
            ('cont_pk', 'int'),
        ])
        for field_name, field_type in fields.iteritems():
            layer.CreateField(
                ogr.FieldDefn(
                    str(field_name), OGR_FIELD_TYPE_MAP[field_type]
                )
            )
        _definition = layer.GetLayerDefn()
        points = reshape_flat_array(selection['coordinates']).T

        for i in xrange(selection['id'].size):
            feature = ogr.Feature(_definition)
            point = ogr.Geometry(ogr.wkbPoint)
            point.AddPoint(points[i][0], points[i][1])
            feature.SetGeometry(point)
            feature.SetField(str("link_id"), int(selection['levl'][i]))
            kcu = selection['kcu'][i]
            feature.SetField(str("kcu"), int(kcu))
            kcu_descr = ''
            try:
                kcu_descr = kcu_dict[kcu]
            except KeyError:
                pass
            feature.SetField(str("kcu_descr"), str(kcu_descr))
            feature.SetField(str("cont_pk"), int(selection['content_pk'][i]))
            layer.CreateFeature(feature)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

import logging

import h5py
import numpy as np

from threedigrid.orm.utils import transform_bbox

from threedigrid.admin.lines.models import Lines
from threedigrid.admin.nodes.models import Nodes
from threedigrid.admin.nodes.models import Cells
from threedigrid.admin.nodes.models import Grid
from threedigrid.admin.breaches.models import Breaches
from threedigrid.admin.pumps.models import Pumps
from threedigrid.admin.levees.models import Levees
from threedigrid.admin.h5py_datasource import H5pyGroup
import constants

logger = logging.getLogger(__name__)


class GridH5Admin(object):

    def __init__(self, h5_file_path, file_modus='r'):

        self.grid_file = h5_file_path
        self.h5py_file = h5py.File(h5_file_path, file_modus)
        self._set_props()

        self._grid_kwargs = {
            'has_1d': self.has_1d
        }

    @property
    def grid(self):
        kwargs = self._grid_kwargs.copy()
        kwargs['n2dtot'] = self.get_from_meta('n2dtot')
        kwargs['dx'] = self.h5py_file['grid_coordinate_attributes']['dx'].value
        return Grid(
            H5pyGroup(self.h5py_file, 'grid_coordinate_attributes'), **kwargs
        )

    @property
    def levees(self):
        return Levees(
            H5pyGroup(self.h5py_file, 'levees'), **self._grid_kwargs)

    @property
    def nodes(self):
        return Nodes(
            H5pyGroup(self.h5py_file, 'nodes'), **self._grid_kwargs)

    @property
    def lines(self):
        return Lines(
            H5pyGroup(self.h5py_file, 'lines'), **self._grid_kwargs)

    @property
    def pumps(self):
        return Pumps(
            H5pyGroup(self.h5py_file, 'pumps'), **self._grid_kwargs)

    @property
    def breaches(self):
        return Breaches(
            H5pyGroup(self.h5py_file, 'breaches'), **self._grid_kwargs)

    @property
    def cells(self):
        # treated as nodes
        return Cells(
            H5pyGroup(self.h5py_file, 'nodes'), **self._grid_kwargs)

    @property
    def revision_hash(self):
        return self.h5py_file.attrs['revision_hash']

    @property
    def revision_nr(self):
        return self.h5py_file.attrs['revision_nr']

    @property
    def model_name(self):
        return self.h5py_file.attrs['model_name']

    @property
    def epsg_code(self):
        return self.h5py_file.attrs['epsg_code']

    def get_extent_subset(self, subset_name, target_epsg_code=''):
        """

        :param subset_name: name of the node subset for which the extent
            should be calculated.
            Valid input: '1D_all', '2D_groundwater'
        :param target_epsg_code: string representation of the desired
            output epsg code
        :returns numpy array of xy-min/xy-max pairs
        """
        attr_name = constants.SUBSET_NAME_H5_ATTR_MAP.get(subset_name.upper())
        extent = self.h5py_file.attrs.get(attr_name, None)
        if extent is None:
            raise AttributeError(
                "The grid admin file %s does not have %s attribute" % (
                    self.grid_file, attr_name)
            )
        if attr_name == constants.EXTENT_1D_KEY and not self.has_1d:
            logger.info(
                '[*] The model has no 1D returning...'
            )
            return
        if attr_name == constants.EXTENT_2D_KEY and not self.has_2d:
            logger.info(
                '[*] The model has no 2D returning...'
            )
            return

        if target_epsg_code and target_epsg_code != self.epsg_code:
            extent = transform_bbox(
                extent, self.epsg_code, target_epsg_code,
            )
        return extent

    def get_model_extent(self, target_epsg_code='', **kwargs):
        """
        get the extent of the model. Combines the extent of 1d and 2d node
        coordinates. If target_epsg_code is different from the models epsg
        code it projects the coordinates to projection of the given
        target epsg code.

        :param target_epsg_code: string representation of the desired
            output epsg code
        :param kwargs:
            extra_extent: list of xy-min/xy-max pairs that need to be
            taken into account in the model extent calculation
        :returns numpy array of xy-min/xy-max pairs
        """
        bbox = kwargs.get('extra_extent', [])
        for k in constants.SUBSET_NAME_H5_ATTR_MAP.keys():
            sub_extent = self.get_extent_subset(
                subset_name=k, target_epsg_code=target_epsg_code)
            if sub_extent is None:
                continue
            bbox.append(sub_extent)

        x = np.array(bbox)[:, [0, 2]]
        y = np.array(bbox)[:, [1, 3]]

        return np.array([np.min(x), np.min(y), np.max(x), np.max(y)])

    @property
    def model_slug(self):
        return self.h5py_file.attrs['model_slug']

    @property
    def has_groundwater(self):
        # TODO threedicore will provide info, change accordingly
        return self.nodes.has_groundwater

    def _set_props(self):
        for prop, value in self.h5py_file.attrs.iteritems():
            if prop and prop.startswith('has_'):
                setattr(self, prop, bool(value))

    @property
    def has_levees(self):
        return hasattr(self, "levees")

    @property
    def threedicore_version(self):
        return self.h5py_file.attrs['threedicore_version']

    def get_from_meta(self, prop_name):
        if prop_name not in self.h5py_file['meta'].keys():
            return None

        return self.h5py_file['meta'][prop_name].value

    def __enter__(self):
        return self.h5py_file

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        self.h5py_file.flush()
        self.h5py_file.close()



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

import ogr

from threedigrid.orm.models import Model
from threedigrid.orm.fields import MultiLineArrayField
from threedigrid.orm.fields import ArrayField
from threedigrid.admin.utils import reshape_flat_array

from threedigrid.admin.levees import exporters


class Levees(Model):

    crest_level = ArrayField()
    max_breach_depth = ArrayField()
    coords = MultiLineArrayField()

    def __init__(self, *args,  **kwargs):
        super(Levees, self).__init__(*args, **kwargs)

        self._geoms = []
        self.current_epsg = None
        self._exporters = [
            exporters.LeveeOgrExporter(self),
        ]

    @property
    def geoms(self):
        if not self._geoms:
            self.load_geoms()
        return self._geoms

    def load_geoms(self):

        if self._geoms:
            return
        for line_coords in self.coords:
            line = ogr.Geometry(ogr.wkbLineString)
            linepoints = reshape_flat_array(line_coords).T
            for x in linepoints:
                line.AddPoint(x[0], x[1])
            self._geoms.append(line)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

SHP_DRIVER_NAME = 'ESRI Shapefile'
GEO_PACKAGE_DRIVER_NAME = 'GPKG'

SHP_EXTENSION = '.shp'
GEO_PACKAGE_EXTENSION = '.gpkg'

EXTENSION_TO_DRIVER_MAP = {
    SHP_EXTENSION: SHP_DRIVER_NAME,
    GEO_PACKAGE_EXTENSION: GEO_PACKAGE_DRIVER_NAME
}



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

import logging

import mercantile
import numpy as np
from pyproj import transform, Proj
import osr

logger = logging.getLogger(__name__)


def angle_in_degrees(x0, y0, x1, y1):
    """
    Calculate the angle of lines made up by the points (p1, p2):

        p1[i] = ([x0[i], y0[i]])
        p2[i] = ([x1[i], y1[i]])

    param x0: np.array with x coordinates of point p1[i]
    param y0: np.array with y coordinates of point p1[i]
    param x1: np.array with x coordinates of point p2[i]
    param y1: np.array with y coordinates of point p2[i]

    Returns: the angles of the lines with the (horizontal) x-axis
             in degrees

    Example:
    >>> angle_in_degrees(np.array([0.0]), np.array([0.0]),
    ...                  np.array([1.0]), np.array([1.0]))
    array([ 45.])
    """

    assert isinstance(x0, np.ndarray)
    assert isinstance(y0, np.ndarray)
    assert isinstance(x1, np.ndarray)
    assert isinstance(y1, np.ndarray)

    # Calculate the differences
    dx, dy = x1 - x0, y1 - y0

    # Divide by zero (dx) solution:
    #   skip (dy/dx)[i] where dx[i] == 0 and replace
    #   the (dy/dx)[i] value with np.inf when dy >= 0 and
    #   -np.inf when dy < 0

    # Create an array with the size of dy with np.inf where
    # dy >=0 and -np.inf where dy < 0
    out_inf = np.full_like(dy, np.inf)
    out_inf[dy < 0] = -np.inf

    # compute dy/dx, but use the out_inf[i] value if dx[i]==0
    dy_divided_by_dx_with_inf = np.divide(
        dy, dx, out=out_inf, where=dx != 0)

    # Return the values of the inverse-tan in degrees
    # np.inf gives 90 and -np.inf gives -90
    return np.degrees(np.arctan(dy_divided_by_dx_with_inf))


def transform_xys(x_array, y_array, source_epsg, target_epsg):
    """
    Transform x_array, y_array from source_epsg_code to
    target_epsg code
    """

    assert isinstance(x_array, np.ndarray)
    assert isinstance(y_array, np.ndarray)

    projections = []
    for epsg_code in (source_epsg, target_epsg):
        epsg_str = 'epsg:{}'.format(epsg_code)
        projection = Proj(init=epsg_str)
        projections.append(projection)

    return np.array(transform(
        projections[0], projections[1], x_array, y_array))


def get_spatial_reference(epsg_code):
    """
    :param epsg_code: Spatial Reference System Identifier (SRID)
        as implemented by the European Petroleum Survey Group (EPSG).

    :return: an osr spatial reference instance

    :raise RuntimeError: if the import from epsg_code failed

    """
    sr = osr.SpatialReference()
    try:
        sr.ImportFromEPSG(int(epsg_code))
        return sr
    except RuntimeError:
        logger.exception(
            "[-] Importing projection from epsg code %s failed." % epsg_code
        )
        raise


MERCANTILE_EPSG_CODE = '4326'

BBOX_LEFT = 0
BBOX_TOP = 1
BBOX_RIGHT = 2
BBOX_BOTTOM = 3


def transform_bbox(bbox, source_epsg_code, target_epsg_code):
    """
    Transform bbox from source_epsg_code to target_epsg_code,
    if necessary
    """
    if source_epsg_code != target_epsg_code:
        # XXX: Not entirely sure whether transformations between two projected
        # coordinate systems always do retain the rectangular shape of a bbox.
        # Transformations between an unprojected system (e.g. WGS84) and a
        # projected system (e.g. RDNEW) will experience distortion: the
        # resulting shape cannot be accurately represented by top left
        # and bottom right.
        source_srs = get_spatial_reference(source_epsg_code)
        target_srs = get_spatial_reference(target_epsg_code)
        if source_srs.IsProjected() != target_srs.IsProjected():
            msg = "Transforming a bbox from %s to %s is inaccurate."
            logger.warning(msg, source_epsg_code, target_epsg_code)
        # Transform to [[left, right],[top, bottom]]
        bbox_trans = np.array(
                transform_xys(
                    np.array([bbox[BBOX_LEFT], bbox[BBOX_RIGHT]]),
                    np.array([bbox[BBOX_TOP], bbox[BBOX_BOTTOM]]),
                    source_epsg_code, target_epsg_code))

        # Transform back to [left,top,right, bottom]
        bbox = np.array(
            [min(bbox_trans[0]),
             min(bbox_trans[1]),
             max(bbox_trans[0]),
             max(bbox_trans[1])])
    return bbox


def get_bbox_for_tile(tile_xyz=(0, 0, 0), target_epsg_code='4326'):
    """
    Get the bbox for a tile defined by x,y,z in epsg=target_epsg_code

    :param tile_xyz: tuple with tile x,y,z
    :param target_epsg_code: the epsg_code to reproject the bbox to.

    :return: The bbox of the tile
             (reprojected to target_epsg_code if necessary)
             Numpy array: [left, top, right, bottom]
             y-axis is inverted: bottom >= top

    Examples:
    >>> get_bbox_for_tile((10, 10, 15))
    array([-179.89013672,   85.04069252, -179.87915039,   85.04164217])

    >>> get_bbox_for_tile((10, 10, 15), '28992')
    array([  212572.58551456,  5468921.2451706 ,   212705.73721283,
            5469031.60724307])
    """
    bbox = np.array(
        mercantile.bounds(
            mercantile.Tile(
                x=int(tile_xyz[0]), y=int(tile_xyz[1]), z=int(tile_xyz[2])))
        )

    if MERCANTILE_EPSG_CODE != target_epsg_code:
        bbox = transform_bbox(bbox, MERCANTILE_EPSG_CODE, target_epsg_code)

    return bbox


def select_points_by_tile(tile_xyz=(0, 0, 0), target_epsg_code='4326',
                          points=None):
    """
    Select points by a tile

    :param tile_xyz: tuple with tile x,y,z
    :param target_epsg_code: the epsg_code to reproject the bbox to
            before selecting.
    :param points: (np.array) array [[x1, x2, x3...], [y1, y2, y3...]]
    :return: A boolean mask for the points within the tile_xyz bbox.

    Example:
    >>> select_points_by_tile((10, 10, 15),
    ...     '4326',
    ...     np.array([[-179.89013670, -179.89013670],
    ...               [  85.04069252,   85.04269252]]))
    array([ True, False], dtype=bool)
    """

    assert isinstance(points, np.ndarray)

    return select_points_by_bbox(
        points, get_bbox_for_tile(tile_xyz, target_epsg_code))


def select_lines_by_tile(tile_xyz=(0, 0, 0), target_epsg_code='4326',
                         lines=None, include_intersections=False):
    """
    Select lines by a tile

    :param tile_xyz: tuple with tile x,y,z
    :param target_epsg_code: the epsg_code to reproject the bbox to
            before selecting.
    :param lines: (np.array) array:
                    x1_array=lines[0], y1_array=lines[1],
                    x2_array=lines[2], y2_array=lines[3]
           line[i] = (x1[i], y1[i]) - (x2[i], y2[i])

    :return: The bbox of the tile
             (reprojected to target_epsg_code if necessary)
    """

    assert isinstance(lines, np.ndarray)

    return select_lines_by_bbox(
        lines, get_bbox_for_tile(tile_xyz, target_epsg_code),
        include_intersections)


def select_points_by_bbox(points, bbox):
    """
    Return a boolean mask array for the points that are in bbox

    :param points: np.array [[x1, x2, x3...], [y1, y2, y3...]]
    :param bbox: np.array [left, top, right, bottom], bottom >= top
    :return: a boolean mask array with 'True' for the
             points that are in the bbox

    Example:
    >>> select_points_by_bbox(
    ...     points=np.array([[1, 10, 1, 10], [10, 10, 50, 50]]),
    ...     bbox=np.array([10, 10, 50, 50]))
    array([False,  True, False,  True], dtype=bool)

    >>> select_points_by_bbox(
    ...     points=np.array([[-11, 10, 10, -10], [10, 10, -11, -10]]),
    ...     bbox=np.array([-10, -10, 10, 10]))
    array([False,  True, False,  True], dtype=bool)

    >>> select_points_by_bbox(
    ...     points=np.array([[-11, 10, 10, -10], [10, 10, 0, -10]]),
    ...     bbox=np.array([-10, 0, 10, 10]))
    array([False,  True,  True, False], dtype=bool)

    >>> select_points_by_bbox(
    ...     points=np.array([[-1, 0, 0, 10], [0, 0, -11, -10]]),
    ...     bbox=np.array([0, -10, 10, 0]))
    array([False,  True, False,  True], dtype=bool)

    >>> select_points_by_bbox(
    ...     points=np.array([[100, -100], [110, -110]]),
    ...     bbox=np.array([0, -10, 10, 0]))
    array([False, False], dtype=bool)
    """
    lleft = np.array([bbox[BBOX_LEFT], bbox[BBOX_TOP]])  # lower-left
    uright = np.array([bbox[BBOX_RIGHT], bbox[BBOX_BOTTOM]])  # upper-right
    xy_points = points.T
    pre_sel_low = xy_points >= lleft
    pre_sel_up = xy_points <= uright
    return np.all((pre_sel_low & pre_sel_up), axis=1)


def lines_to_bbox_array(lines):
    """
    Convert lines:
        x1_array=lines[0], y1_array=lines[1],
        x2_array=lines[2], y2_array=lines[3]

    to bbox_array:
        bbox[i] = [left, top, right, bottom]

    :param lines:
        x1_array=lines[0], y1_array=lines[1],
        x2_array=lines[2], y2_array=lines[3]

    :return: np.array with bboxes
    """
    # Use min/max, don't assume the x1[i] to always be left,
    # y1[i] to be top etc..
    return np.stack(
        (np.fmin(lines[0], lines[2]), np.fmin(lines[1], lines[3]),
         np.fmax(lines[0], lines[2]), np.fmax(lines[1], lines[3])), axis=1)


def select_lines_by_bbox(lines, bbox, include_intersections=False):
    """
    Return a boolean mask array for the lines that are in bbox

    :param lines: np.array
                x1_array=lines[0], y1_array=lines[1],
                x2_array=lines[2], y2_array=lines[3]

                line[i] = (x1[i], y1[i]) - (x2[i], y2[i])

    :param bbox: np.array [left, top, right, bottom]
    :return: a boolean mask array with 'True' for the
             points that are in the bbox

    """
    assert isinstance(lines, np.ndarray)

    # We can treat lines same like bboxes
    # if we transform the lines array's to bboxes
    bbox_array = lines_to_bbox_array(lines)

    if include_intersections:
        return get_bbox_intersect_bool_mask(bbox_array, bbox)

    return get_bbox_in_bool_mask(bbox_array, bbox)


def get_bbox_in_bool_mask(bbox_array, bbox):
    """
    Return a boolean mask with 'True' when bbox_array[i] is in bbox2

    :param bbox_array: np.array with bboxes [left, top, right, bottom]
    :param bbox: np.array to match with bbox_array items
                 [left, top, right, bottom]
    :return: A boolean mask with 'True' when bbox_array[i] is in
             with bbox

    Example:
    >>> get_bbox_in_bool_mask(
    ...     bbox_array=np.array(
    ...         [[5, 5, 10, 10], [15, 15, 30, 30], [50, 50, 60, 60]]),
    ...     bbox=np.array([0, 0, 20, 20]))
    array([ True, False, False], dtype=bool)
    """

    assert isinstance(bbox_array, np.ndarray)

    return ((bbox[BBOX_LEFT] <= bbox_array[:, BBOX_LEFT]) &
            (bbox[BBOX_TOP] <= bbox_array[:, BBOX_TOP]) &
            (bbox[BBOX_RIGHT] >= bbox_array[:, BBOX_RIGHT]) &
            (bbox[BBOX_BOTTOM] >= bbox_array[:, BBOX_BOTTOM])).flatten()


def get_bbox_by_point(pnt, values):
    """

    :param pnt: list or tuple of xy values
    :param values: arrays of xy's of the lower left and
        upper right corner, e.g. [[x,...], [y,...], [x1,...], [y1,...]]]
    :returns the bbox (lower left and upper right corner)
        in values which contains pnt
    Example:
    >>> get_bbox_by_point(
    ...     pnt=(1,5),
    ...     values=np.array([[0,2], [0,2], [6,6], [10,11]]))
    array([ True, False], dtype=bool)
    """
    if isinstance(pnt, (list, tuple)):
        pnt = np.array(pnt)

    assert isinstance(pnt, np.ndarray)
    # pnt --> np.array(x, y)
    lleft = np.array((values[0], values[1]))
    uright = np.array((values[2], values[3]))

    pre_sel_low = pnt >= lleft.T
    pre_sel_up = pnt <= uright.T
    return np.all((pre_sel_low & pre_sel_up), axis=1)


def get_bbox_intersect_bool_mask(bbox_array, bbox):
    """
    Return a boolean mask with 'True' when bbox_array[i] intersects with bbox2

    :param bbox_array: np.array with bboxes [left, top, right, bottom]
    :param bbox: np.array to match with bbox_array items
                 [left, top, right, bottom]
    :return: A boolean mask with 'True' when bbox_array[i] intersects
             with bbox

    Example:
    >>> get_bbox_intersect_bool_mask(
    ...     bbox_array=np.array(
    ...         [[5, 5, 10, 10], [15, 15, 30, 30], [50, 50, 60, 60]]),
    ...     bbox=np.array([0, 0, 20, 20]))
    array([ True,  True, False], dtype=bool)
    """

    assert isinstance(bbox_array, np.ndarray)

    # Check which ones do not intersect with:
    #    bbox[BBOX_LEFT] > bbox_array[:, BBOX_RIGHT] etc..
    #
    #              bbox_array[i, BBOX_RIGHT]
    #              V
    #          ----- -----
    #          |   | |   |
    #          ----- -----
    #                ^
    #                bbox[BBOX_LEFT]
    #
    # Invert the boolean array and flatten.
    return np.invert(
            (bbox[BBOX_LEFT] > bbox_array[:, BBOX_RIGHT]) |
            (bbox[BBOX_RIGHT] < bbox_array[:, BBOX_LEFT]) |
            (bbox[BBOX_TOP] > bbox_array[:, BBOX_BOTTOM]) |
            (bbox[BBOX_BOTTOM] < bbox_array[:, BBOX_TOP])).flatten()


def get_bbox_intersections(bbox_array, bbox):
    """
    Return the intersections and indices of bbox_array that intersect
    with bbox2

    :param bbox_array: np.array with bboxes [left, top, right, bottom]
    :param bbox: np.array to match with bbox_array items
                 [left, top, right, bottom]
    :return: the intersections and indices of bbox_array which intersect with
             bbox2

    Example:
    >>> get_bbox_intersections(
    ...     bbox_array=np.array(
    ...         [[5, 5, 10, 10], [15, 15, 30, 30], [50, 50, 60, 60]]),
    ...     bbox=np.array([0, 0, 20, 20]))
    (array([[ 5,  5, 10, 10],
           [15, 15, 20, 20]]), array([0, 1]))
    """

    assert isinstance(bbox_array, np.ndarray)

    intersect_mask = get_bbox_intersect_bool_mask(bbox_array, bbox)
    nodes_intersect = bbox_array[intersect_mask]

    # Calculate the intersections by getting the:
    #     max values for: left, top
    #     miniminum for: right, bottom
    #
    #     For example (left):
    #
    #          bbox_array[intersection_mask][i, BBOX_LEFT]
    #          V
    #          -----
    #          |   |
    #          -----
    #             -----
    #             |   |
    #             -----
    #             ^
    #             bbox[BBOX_LEFT]
    #
    #      intersection left value is max of both values.
    intersections = np.concatenate(
        (np.fmax(nodes_intersect[:, 0:2], bbox[0:2]),
         np.fmin(nodes_intersect[:, 2:4], bbox[2:4])), axis=1)
    return (intersections, np.argwhere(intersect_mask).flatten())


def reshape_flat_array(flat_array):
    return flat_array.reshape(2, -1)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from threedigrid.admin.breaches import exporters
from threedigrid.admin import constants
from threedigrid.orm.models import Model
from threedigrid.orm.fields import (
    ArrayField, PointArrayField)
from threedigrid.orm.base.fields import IndexArrayField


class Breaches(Model):
    content_pk = ArrayField()
    seq_ids = ArrayField()
    levbr = ArrayField()
    levl = IndexArrayField(to='Lines')
    levmat = ArrayField()
    kcu = ArrayField()

    coordinates = PointArrayField()

    OBJECT_TYPE = constants.TYPE_V2_BREACH

    def __init__(self, *args, **kwargs):
        super(Breaches, self).__init__(*args, **kwargs)

        self._exporters = [
            exporters.BreachesOgrExporter(self),
        ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'Click>=6.0',
    'numpy>=1.13',
    'h5py>=2.7.1',
    'mercantile>=1.0.1',
    'pyproj>=1.9.5.1',
    'Shapely>=1.6.4',
    'geojson>=2.3.0'
]

setup_requirements = []

test_requirements = ['pytest==3.4.1', ]

setup(
    author="Lars Claussen",
    author_email='lars.claussen@nelen-schuurmans.nl',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
    ],
    description="Python package for the threedigrid administration",
    entry_points={
        'console_scripts': [
            'threedigrid=threedigrid.cli:main',
        ],
    },
    install_requires=requirements,
    license="BSD license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='threedigrid',
    name='threedigrid',
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/nens/threedigrid',
    version='0.1.2',
    zip_safe=False,
)
